package de.fhmuenster.masterthesis.generator;

import java.util.List;

import de.fhmuenster.masterthesis.generator.mockdata.ExternalServiceMock;
import de.fhmuenster.masterthesis.generator.mockdata.MockCompare;
import de.fhmuenster.masterthesis.generator.mockdata.MockDeclaration;
import de.fhmuenster.masterthesis.generator.mockdata.MockRule;
import de.fhmuenster.masterthesis.generator.mockdata.ObjectCompare;
import de.fhmuenster.masterthesis.generator.mockdata.PrimitiveCompare;

public class ExternalServiceMockGenerator extends AbstractTestgenerator {

	private static final String CLASS_DECLARATION = "public class ExternalServiceMockProvider {";
	private static final String VARIABLE_DECLARATION = "\tprivate static final String %s = \"%s\";";
	private static final String MOCK_METHOD_HEADER = "\tpublic static final Map<String, Object> %s(Map<String, Object> input) {";
	private static final String STRING_VARIABLE_READ = "\t\tString %s = (String) input.get(%s);";
	private static final String INT_VARIABLE_READ = "\t\tint %s = (int) input.get(%s);";
	private static final String BOOLEAN_VARIABLE_READ = "\t\tboolean %s = (boolean) input.get(%s);";
	
	private StringBuilder codegeneratorMock = new StringBuilder();

	private List<ExternalServiceMock> externalServiceMocks;
	private String packageName;
	
	public ExternalServiceMockGenerator(List<ExternalServiceMock> externalServiceMocks, String packageName) {
		this.externalServiceMocks = externalServiceMocks;
		this.packageName = packageName;
	}

	public void generateMock() {
		addPackageDeclaration(packageName, codegeneratorMock);
		addSeparatorLine(codegeneratorMock);
		
		addImports();
		addSeparatorLine(codegeneratorMock);
		
		addExternalServiceClassHeader();
		addSeparatorLine(codegeneratorMock);

		addVariableDeclarations();
		addSeparatorLine(codegeneratorMock);
		
		int counter = 0;
		for (ExternalServiceMock mock : externalServiceMocks) {
			if(counter > 0) {
				addSeparatorLine(codegeneratorMock);
			}
			
			addMethodHeader(mock.getName());
			
			addVariableReads(mock);
			addSeparatorLine(codegeneratorMock);
		
			addMockRules(mock);
			addSeparatorLine(codegeneratorMock);
			
			addMethodClosing();
			
			counter++;
		}

		addClassClosing(codegeneratorMock);
	}

	private void addImports() {
		appendLine("import java.util.HashMap;", codegeneratorMock);
		appendLine("import java.util.Map;", codegeneratorMock);
	}
	
	private void addExternalServiceClassHeader() {
		appendLine(CLASS_DECLARATION, codegeneratorMock);
	}
	
	private void addVariableDeclarations() {
		externalServiceMocks.stream() //
			.flatMap(mock -> mock.getVariables().stream()) //
			.distinct() //
			.forEach(variable -> addVariableDeclaration(variable)); //
	}

	private void addVariableDeclaration(String variableName) {
		appendLine(String.format(VARIABLE_DECLARATION, NameProvider.getStaticFinalVariableName(variableName), variableName), codegeneratorMock);
	}
	
	private void addMethodHeader(String mockName) {
		appendLine(String.format(MOCK_METHOD_HEADER, mockName), codegeneratorMock);
		appendLine("\t\tMap<String, Object> output = new HashMap<String, Object>();", codegeneratorMock);
		addSeparatorLine(codegeneratorMock);
	}
	
	private void addVariableReads(ExternalServiceMock mock) {
		for(String stringVariableName : mock.getReadVariables(String.class)) {
			String variableNameReference = NameProvider.getStaticFinalVariableName(stringVariableName);
			appendLine(String.format(STRING_VARIABLE_READ, stringVariableName, variableNameReference), codegeneratorMock);
		}
		
		for(String intVariableName : mock.getReadVariables(Integer.class)) {
			String variableNameReference = NameProvider.getStaticFinalVariableName(intVariableName);
			appendLine(String.format(INT_VARIABLE_READ, intVariableName, variableNameReference), codegeneratorMock);
		}
		
		for(String booleanVariableName : mock.getReadVariables(Boolean.class)) {
			String variableNameReference = NameProvider.getStaticFinalVariableName(booleanVariableName);
			appendLine(String.format(BOOLEAN_VARIABLE_READ, booleanVariableName, variableNameReference), codegeneratorMock);
		}
	}
	
	private void addMockRules(ExternalServiceMock mock) {
		int counter = 0;
		int amountMockRules = mock.getMockRules().size();
		
		for (MockRule mockRule : mock.getMockRules()) {
			String compareStatement = getCompareStatement(mockRule.getComparisons());
			
			if(counter == 0) {
				appendLine(String.format("\t\tif(%s) {", compareStatement), codegeneratorMock);
			} else if(!"".equals(compareStatement)){
				appendLine(String.format("\t\t} else if(%s) {", compareStatement), codegeneratorMock);
			} else if("".equals(compareStatement) && (counter == amountMockRules - 1)) {
				appendLine(String.format("\t\t} else {", compareStatement), codegeneratorMock);
			}
			
			for (MockDeclaration mockDeclaration : mockRule.getDeclarations()) {
				String variableName = NameProvider.getStaticFinalVariableName(mockDeclaration.getVariable());
				String value = getValue(mockDeclaration.getValue());
				
				appendLine(String.format("\t\t\toutput.put(%s, %s);", variableName, value), codegeneratorMock);
			}
			
			if(counter == (amountMockRules - 1)) {
				appendLine("\t\t}", codegeneratorMock);
			}
			
			counter++;
		}
	}
	
	private String getCompareStatement(List<MockCompare> comparisons) {
		String compareStatement = "";
		
		int counter = 0;
		for (MockCompare compare : comparisons) {
			if(counter != 0) {
				compareStatement += " && ";
			}
			
			if(compare instanceof PrimitiveCompare) {
				String variableName = compare.getVariable();
				String compareOperator = ((PrimitiveCompare) compare).getCompareOperator();
				Object value = compare.getValue();
				
				compareStatement += (variableName + " " + compareOperator + " " + value);
			}
			
			if(compare instanceof ObjectCompare) {
				String variableName = compare.getVariable();
				String negation = ((ObjectCompare) compare).isPositiveCheck() ? "" : "!";
				Object value = compare.getValue();
				
				compareStatement += (negation + "\"" + value + "\"" + ".equals(" + variableName + ")");
			}
			
			counter++;
		}
		
		return compareStatement;
	}
	
	private String getValue(Object value) {
		if (value instanceof String) {
			return "\"" + value + "\"";
		}
		return value.toString();
	}
	
	private void addMethodClosing() {
		appendLine("\t\treturn output;", codegeneratorMock);
		appendLine("\t}", codegeneratorMock);
	}
	
	public String getMockCode() {
		return codegeneratorMock.toString();
	}
}
